/*
	Copyright (C) 1995-2001 AROS - The Amiga Research OS
	$Id: windowclasses.c,v 1.18 2009/02/22 21:55:02 jacadcaps Exp $

	Desc: Classes for window decor stuff, like dragbar, close etc.

	Lang: english
*/

/***********************************************************************************/

#include <string.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/utility.h>
#include <proto/layers.h>
#include <proto/dos.h>

#include <intuition/classes.h>
#include <intuition/gadgetclass.h>
#include <intuition/cghooks.h>
#include <intuition/imageclass.h>
#include <devices/input.h>
#include <aros/asmcall.h>

#include "gadgets.h"

#include "intuition_intern.h"
#include "inputhandler.h"
#include "inputhandler_support.h"
#include "inputhandler_actions.h"

#ifdef SKINS
#define NOBUTTONFOCUS (IBase->IControlPrefs.ic_Flags & ICF_NOMOUSEACTIONS)
#define NOPOINTERCHANGE (IBase->IControlPrefs.ic_Flags & ICF_NOPOINTERCHANGE)
#define POINTERCHANGE (!NOBUTTONFOCUS && !NOPOINTERCHANGE)

#include "intuition_customizesupport.h"
#include "renderwindowframe.h"
#include "mosmisc.h"
#define LAYERS52 ((((struct Library *)LayersBase)->lib_Version >= 52) && w && IS(w->WScreen)->Compositing)
#endif

#undef SDEBUG
#undef DEBUG
#define SDEBUG 0
#define DEBUG 0
#include <aros/debug.h>

#define MOVEHACK (IBase->IControlPrefs.ic_Flags & ICF_PRIVILEDGEDREFRESH)

/*

jDc: opaque resize code is NOT sutiable for other apps than MUI, I have no intentions in fixing it so that
it would work with all apps out there. Use mui or die :)

The idea behind the trick is that we use a func that checks if apps reply to idcmps in some period
of time. As long as they are replying fast we can do another layer resize, send idcmp to app.
If they do not then we wait for a while and after some time, if one or more apps don't reply
we do the resize again (the timeout is also useful when we have some dead app with windows
on screen.

*/

#ifdef __MORPHOS__
#define OPAQUESIZE ((w->MoreFlags & WMFLG_IAMMUI) && (IBase->IControlPrefs.ic_Flags & ICF_OPAQUERESIZE))
#else
#define OPAQUESIZE 0
#endif

#define DELAYEDDRAG
#define DELAYEDSIZE

/***********************************************************************************/

#define G(o)  ((struct Gadget *)o)
#define EG(o) ((struct ExtGadget *)o)
#define IM(o) ((struct Image *)o)

/***********************************************************************************/

struct dragbar_data
{
	/* Current left- and topedge of moving window. Ie when the user releases
	the LMB after a windowdrag, the window's new coords will be (curleft, curtop)
	*/

	LONG curleft;
	LONG curtop;

	/* The current x and y coordinates relative to curleft/curtop */
	LONG mousex;
	LONG mousey;

	/* Whether the dragframe is currently drawn or erased */
	BOOL isrendered;

	/* Used to tell GM_GOINACTIVE whether the drag was canceled or not */
	BOOL drag_canceled;

	/* Used to tell GM_GOINACTIVE whether UnlockLayer() or not */
	BOOL drag_layerlock;
#ifdef USEGADGETLOCK
 	 BOOL drag_gadgetlock;
#endif
	BOOL drag_refreshed;

	/* Rastport to use during update */
	struct RastPort *rp;

	UQUAD lasteventtime;
	
	LONG startleft;
	LONG starttop;

#ifdef USEWINDOWLOCK
	/* used to prevent windows from opening/closing while user drags a window */
	BOOL drag_windowlock;
#endif

	struct Task *movetask;

	BOOL drag_newtimer;

};

/***********************************************************************************/

#define WSIG_MOVE SIGF_INTUITION
#define WSIG_DIE  SIGF_ABORT

/***********************************************************************************/

void MoveTask(struct dragbar_data *data,struct Window *w,struct Screen *screen,struct IntuitionBase *IntuitionBase)
{
	ULONG signals;

	for (;;)
	{
		signals = Wait(WSIG_DIE|WSIG_MOVE);

		if (signals & WSIG_DIE)
		{
			Forbid();
			data->movetask = 0;
			Permit();
			WindowAction(w,WAC_CHANGEWINDOWBOX,0);
			return;
		}

		if (signals & WSIG_MOVE)
		{
			struct Layer *L;

			LockLayers(&screen->LayerInfo);

			for (L = screen->LayerInfo.top_layer; L; L = L->back)
			{
				if (L->Window == w)
				{
					//window not closed yet!
					if (data->curleft != w->LeftEdge || data->curtop != w->TopEdge)
					{
						struct Requester	*req;

						if (BLAYER(w))
						{
							/* move outer window first */
							MoveSizeLayer(BLAYER(w), data->curleft - w->LeftEdge, data->curtop - w->TopEdge, 0, 0);
						}

						MoveSizeLayer(WLAYER(w), data->curleft - w->LeftEdge, data->curtop - w->TopEdge, 0, 0);

						for (req = w->FirstRequest; req; req = req->OlderRequest)
						{
							struct Layer *layer = req->ReqLayer;

							if (layer)
							{
								int dx, dy, dw, dh;
								int left, top, right, bottom;

								left = data->curleft + req->LeftEdge;
								top = data->curtop + req->TopEdge;
								right = left + req->Width - 1;
								bottom = top + req->Height - 1;

								if (left > data->curleft + w->Width - 1)
									left = data->curleft + w->Width - 1;

								if (top > data->curtop + w->Height - 1)
									top = data->curtop + w->Height - 1;

								if (right > data->curleft + w->Width - 1)
									right = data->curleft + w->Width - 1;

								if (bottom > data->curtop + w->Height - 1)
									bottom = data->curtop + w->Height - 1;

								dx = left - layer->bounds.MinX;
								dy = top - layer->bounds.MinY;
								dw = right - left - layer->bounds.MaxX + layer->bounds.MinX;
								dh = bottom - top - layer->bounds.MaxY + layer->bounds.MinY;

								MoveSizeLayer(layer, dx, dy, dw, dh);
							}
						}

						w->LeftEdge = data->curleft;
						w->TopEdge = data->curtop;
						UpdateMouseCoords(w);

						if (screen->Flags & AUTOSCROLL) ScreenPosition(screen,SPOS_MAKEVISIBLE,w->LeftEdge,w->TopEdge,w->LeftEdge + w->Width - 1,w->TopEdge + w->Height - 1);

					}
				}
			}

			#ifdef DAMAGECACHE
			RecordDamage(screen,IntuitionBase);
			#endif
			data->drag_refreshed = FALSE;

			UnlockLayers(&screen->LayerInfo);
		}
	}
}

/***********************************************************************************/

#define IntuitionBase ((struct IntuitionBase *)(cl)->cl_UserData)

/***********************************************************************************/

#if 0
static VOID dragbar_render(Class *cl, Object *o, struct gpRender * msg)
{
	EnterFunc(bug("DragBar::Render()\n"));
	/* We will let the AROS gadgetclass test if it is safe to render */
	if ( DoSuperMethodA(cl, o, (Msg)msg) != 0)
	{
		struct DrawInfo		*dri = msg->gpr_GInfo->gi_DrInfo;
		UWORD			*pens = dri->dri_Pens;
		struct RastPort		*rp = msg->gpr_RPort;
		struct IBox		container;
		struct Window		*win = msg->gpr_GInfo->gi_Window;
		struct TextExtent	te;

		GetGadgetIBox(o, msg->gpr_GInfo, &container);

		if (container.Width <= 1 || container.Height <= 1)
			return;


		/* Clear the dragbar */

		SetAPen(rp, (win->Flags & WFLG_WINDOWACTIVE) ?
			pens[FILLPEN] : pens[BACKGROUNDPEN]);

		SetDrMd(rp, JAM1);

		D(bug("Filling from (%d, %d) to (%d, %d)\n",
		      container.Left,
		      container.Top,
		      container.Left + container.Width - 1,
		      container.Top + container.Height - 1));

		RectFill(rp,
			 container.Left,
			 container.Top,
			 container.Left + container.Width - 1,
			 container.Top + container.Height - 2);

		/* Draw a thin dark line around the bar */

		SetAPen(rp, pens[SHINEPEN]);
		RectFill(rp,container.Left,
			 container.Top,
			 container.Left,
			 container.Top + container.Height - 1 - ((container.Left == 0) ? 0 : 1));
		RectFill(rp,container.Left + 1,
			 container.Top,
			 container.Left + container.Width - 1,
			 container.Top);

		SetAPen(rp,pens[SHADOWPEN]);
		RectFill(rp,container.Left + container.Width - 1,
			 container.Top + 1,
			 container.Left + container.Width - 1,
			 container.Top + container.Height - 1);
		RectFill(rp,container.Left + ((container.Left == 0) ? 1 : 0),
			 container.Top + container.Height - 1,
			 container.Left + container.Width - 2,
			 container.Top + container.Height - 1);

		/* Render the titlebar */
		if (NULL != win->Title)
		{
			ULONG textlen, titlelen;

			SetFont(rp, dri->dri_Font);

			titlelen = strlen(win->Title);
			textlen = TextFit(rp
					  , win->Title
					  , titlelen
					  , &te
					  , NULL
					  , 1
					  , container.Width - 6
					  , container.Height);

			SetAPen(rp, pens[(win->Flags & WFLG_WINDOWACTIVE) ? FILLTEXTPEN : TEXTPEN]);
			Move(rp, container.Left + 3, container.Top + dri->dri_Font->tf_Baseline + 2);

			Text(rp, win->Title, textlen);
		}

	}  /* if (allowed to render) */

	ReturnVoid("DragBar::Render");
}
#endif

/***********************************************************************************/

static IPTR dragbar_goactive(Class *cl, Object *o, struct gpInput *msg)
{

	IPTR retval = GMR_NOREUSE;

	struct InputEvent *ie = msg->gpi_IEvent;

	if (ie)
	{
		/* The gadget was activated via mouse input */
		struct dragbar_data	*data;
		struct Window		*w;

		/* There is no point in rerendering ourseleves her, as this
		   is done by a call to RefreshWindowFrame() in the intuition inputhandler
		*/

		w = msg->gpi_GInfo->gi_Window;

		data = INST_DATA(cl, o);

		data->curleft = w->LeftEdge;
		data->curtop  = w->TopEdge;
		
		data->startleft = w->LeftEdge;
		data->starttop = w->TopEdge;

		data->mousex = w->WScreen->MouseX - data->curleft;
		data->mousey = w->WScreen->MouseY - data->curtop;

		data->drag_refreshed = TRUE;

#ifdef USEWINDOWLOCK
		/* do NOT ObtainSemaphore here since this would lead to deadlocks!!! */
		/* when the semaphore can not be obtained we simply ignore gadget activation */

		/* in movehack the task that locks windowlock is the one that calls MoveWindow*/
		if (!(ATTEMPTWINDOWLOCK))
		{
			goto fail;
		}
		data->drag_windowlock = TRUE;
#endif

		data->rp = CloneRastPort(&w->WScreen->RastPort);

		if (data->rp)
		{

			/* Lock all layers while the window is dragged.
			 * Get the gadget lock first to avoid deadlocks
			 * with ObtainGIRPort. */

			D(bug("locking all layers\n"));

			if (!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE))
			{
#ifdef USEGADGETLOCK
				if (AttemptSemaphore(&IBase->GadgetLock))
				{
					data->drag_gadgetlock = TRUE;
				}
				else
				{
					goto fail;
				}
#endif
#ifndef DELAYEDDRAG
				LockLayers(&w->WScreen->LayerInfo);
				data->drag_layerlock = TRUE;
#endif
			} else {

				/* newtimer on! */
				{
					struct IOStdReq req;
					ULONG cmd = IND_SETNEWTIMER_START;

					memclr(&req,sizeof (struct IOStdReq));

					req.io_Device = IBase->InputIO->io_Device;
					req.io_Unit = IBase->InputIO->io_Unit;
					req.io_Command = IND_SETNEWTIMER;
					req.io_Data = &cmd;
					DoIO((struct IORequest *)&req);

					data->drag_newtimer = TRUE;

				}

				if (MOVEHACK)
				{
	//disabled, use IECLASS_NEWTIMER events for now
/*					  data->movetask =
						NewCreateTask(TASKTAG_CODETYPE, CODETYPE_PPC, TASKTAG_PC, (ULONG)MoveTask,
										TASKTAG_PRI, 0,
										TASKTAG_PPC_ARG1,(ULONG)data,
										TASKTAG_PPC_ARG2,(ULONG)w,
										TASKTAG_PPC_ARG3,(ULONG)w->WScreen,
										TASKTAG_PPC_ARG4,(ULONG)IntuitionBase,
										TAG_DONE);
*/
				}
			}

			SetDrMd(data->rp, COMPLEMENT);

#ifndef DELAYEDDRAG
			if (!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE))
			{
				drawwindowframe(w->WScreen
						, data->rp
						, data->curleft
						, data->curtop
						, data->curleft + w->Width  - 1
						, data->curtop  + w->Height - 1
						, IntuitionBase
					       );

				data->isrendered = TRUE;
			}
#endif
			if (POINTERCHANGE)
			{
				IBase->GadgetPointer = POINTERTYPE_MOVE;
				UpdateScreenPointer(w->WScreen,IntuitionBase);
			}

			data->drag_canceled = FALSE;

			{
				UQUAD currenttime;
				currenttime = ie->ie_TimeStamp.tv_secs;
				currenttime = currenttime * 50;
				currenttime += ie->ie_TimeStamp.tv_micro / 20000;
				data->lasteventtime = currenttime;
			}

			return GMR_MEACTIVE;
		}

fail:

		if (data->drag_layerlock)
		{
			UnlockLayers(&w->WScreen->LayerInfo);
			data->drag_layerlock = FALSE;
			;
		}

#ifdef USEGADGETLOCK
		if (data->drag_gadgetlock)
		{
			ReleaseSemaphore(&IBase->GadgetLock);
			data->drag_gadgetlock = FALSE;
		}
#endif

#ifdef USEWINDOWLOCK
		if (data->drag_windowlock)
		{
			UNLOCKWINDOW;
			data->drag_windowlock = FALSE;
		}
#endif
	}

	return retval;

}

/***********************************************************************************/

static IPTR dragbar_movetest(Class *cl, Object *o, struct gpInput *msg)
{
	IPTR retval = MOVETEST_MOVE;
	struct dragbar_data	*data = INST_DATA(cl, o);
	struct Window		*w = msg->gpi_GInfo->gi_Window;
	struct InputEvent myie;
	LONG		new_left;
	LONG		new_top;

#ifdef SKINS
	CopyMem(msg->gpi_IEvent,&myie,sizeof (struct InputEvent));
	myie.ie_Code = 0x68; //mouse_leftpress

	/* Can we move to the new position, or is window at edge of display ? */
	new_left = msg->gpi_Mouse.X - data->mousex;
	new_top  = msg->gpi_Mouse.Y - data->mousey;

	if (((!(IBase->IControlPrefs.ic_Flags & ICF_OFFSCREENLAYERS)) && !MatchHotkey(&myie,IA_TOGGLEOFFSCREEN,IntuitionBase)) ||
		((IBase->IControlPrefs.ic_Flags & ICF_OFFSCREENLAYERS) && MatchHotkey(&myie,IA_TOGGLEOFFSCREEN,IntuitionBase)))
	{
		if (new_left < 0)
		{
			msg->gpi_Mouse.X -= new_left;
			retval = MOVETEST_ADJUSTPOS;
		}

		if (new_top < 0)
		{
			msg->gpi_Mouse.Y -= new_top;
			retval = MOVETEST_ADJUSTPOS;
		}

		if (new_left + w->Width > w->WScreen->Width - w->WScreen->LeftEdge)
		{
			msg->gpi_Mouse.X -= new_left - (w->WScreen->Width - w->WScreen->LeftEdge - w->Width);
			retval = MOVETEST_ADJUSTPOS;
		}

		if (new_top + w->Height > w->WScreen->Height - w->WScreen->TopEdge)
		{
			msg->gpi_Mouse.Y -= new_top - (w->WScreen->Height - w->WScreen->TopEdge - w->Height);
			retval =  MOVETEST_ADJUSTPOS;
		}
	}
#endif

	return retval;

}

/***********************************************************************************/

static IPTR dragbar_handleinput(Class *cl, Object *o, struct gpInput *msg)
{
	IPTR retval = GMR_MEACTIVE;
	struct GadgetInfo *gi = msg->gpi_GInfo;

	if (gi)
	{
		struct InputEvent	*ie = msg->gpi_IEvent;
		struct dragbar_data	*data = INST_DATA(cl, o);
		struct Window		*w = msg->gpi_GInfo->gi_Window;

		switch (ie->ie_Class)
		{
		case IECLASS_RAWMOUSE:
			switch (ie->ie_Code)
			{
			case MENUDOWN:
				retval = GMR_NOREUSE;
				data->drag_canceled = TRUE;
				break;

			case SELECTUP:
				retval = GMR_NOREUSE;
				break;


			case IECODE_NOBUTTON:
				{
					struct Screen	*scr = w->WScreen;
					struct InputEvent myie;
					LONG new_left;
					LONG new_top;
					//LONG offset = scr->Width >= 1024 ? 10 : 5;

					/* Can we move to the new position, or is window at edge of display ? */
					new_left = scr->MouseX - data->mousex;
					new_top  = scr->MouseY - data->mousey;

					CopyMem(ie,&myie,sizeof (struct InputEvent));
					myie.ie_Code = SELECTDOWN; //mouse_leftpress

					if (((!(IBase->IControlPrefs.ic_Flags & ICF_OFFSCREENLAYERS)) && !MatchHotkey(&myie,IA_TOGGLEOFFSCREEN,IntuitionBase)) ||
					((IBase->IControlPrefs.ic_Flags & ICF_OFFSCREENLAYERS) && MatchHotkey(&myie,IA_TOGGLEOFFSCREEN,IntuitionBase)))
					{
						if (new_left < 0)
						{
							data->mousex += new_left;
							new_left = 0;
						}

						if (new_top < 0)
						{
							data->mousey += new_top;
							new_top = 0;
						}

						if (new_left + w->Width > scr->Width)
						{
							LONG correct_left;
							correct_left = scr->Width - w->Width; /* align to screen border */
							//data->mousex += new_left - correct_left;
							new_left = correct_left;
						}

						if (new_top + w->Height > scr->Height)
						{
							LONG correct_top;
							correct_top = scr->Height - w->Height; /* align to screen border */
							//data->mousey += new_top - correct_top;
							new_top = correct_top;
						}
					}
#if 0
// offset stuff
					else
					{
						if ((new_left < 0) && (new_left > -offset) && (w->LeftEdge >= 0))
						{
							new_left = 0;
						}

						if ((new_top < 0) && (new_top > -offset) && (w->TopEdge >= 0))
						{
							new_top = 0;
						}
					
						if (new_left + w->Width > scr->Width && new_left + w->Width - scr->Width < offset && w->LeftEdge + w->Width <= scr->Width)
						{
							new_left = scr->Width - w->Width;
						}
				
						if (new_top + w->Height > scr->Height && new_top + w->Height - scr->Height < offset && w->TopEdge + w->Height <= scr->Height)
						{
							new_top = scr->Height - w->Height;
						}
					}
#endif

					if (data->curleft != new_left || data->curtop != new_top)
					{
						SetDrMd(data->rp, COMPLEMENT);

						if ((data->isrendered) && (!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE)))
						{
							/* Erase old frame */
							drawwindowframe(w->WScreen
									, data->rp
									, data->curleft
									, data->curtop
									, data->curleft + w->Width  - 1
									, data->curtop  + w->Height - 1
									, IntuitionBase
								       );
						}

						data->curleft = new_left;
						data->curtop  = new_top;

						if (IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE)
						{
							WORD newx = new_left - w->LeftEdge, newy = new_top - w->TopEdge;

							if (newx || newy)
							{
								if (POINTERCHANGE && !data->isrendered)
								{
									IBase->GadgetPointer = POINTERTYPE_MOVE;
									UpdateScreenPointer(w->WScreen,IntuitionBase);
									data->isrendered = TRUE;
								}

								if (data->movetask)
								{
									Forbid();
									Signal(data->movetask,WSIG_MOVE);
									Permit();
									CheckLayers(w->WScreen, IntuitionBase);
								} else {
									if (!(MOVEHACK))
									{
										struct Requester *req;

										ih_fire_intuimessage(   w,
															  IDCMP_CHANGEWINDOW,
															  CWCODE_MOVESIZE,
															  w,
															  IntuitionBase);

										LockLayers(&w->WScreen->LayerInfo);

										w->LeftEdge    += newx;
										w->TopEdge     += newy;

										if (BLAYER(w))
										{
											MoveSizeLayer(BLAYER(w), newx, newy , 0, 0);
										}

										MoveSizeLayer(WLAYER(w), newx, newy, 0, 0);

										for (req = w->FirstRequest; req; req = req->OlderRequest)
										{
											struct Layer *layer = req->ReqLayer;

											if (layer)
											{
												int dx, dy, dw, dh;
												int left, top, right, bottom;

												left = data->curleft + req->LeftEdge;
												top = data->curtop + req->TopEdge;
												right = left + req->Width - 1;
												bottom = top + req->Height - 1;

												if (left > data->curleft + w->Width - 1)
													left = data->curleft + w->Width - 1;

												if (top > data->curtop + w->Height - 1)
													top = data->curtop + w->Height - 1;

												if (right > data->curleft + w->Width - 1)
													right = data->curleft + w->Width - 1;

												if (bottom > data->curtop + w->Height - 1)
													bottom = data->curtop + w->Height - 1;

												dx = left - layer->bounds.MinX;
												dy = top - layer->bounds.MinY;
												dw = right - left - layer->bounds.MaxX + layer->bounds.MinX;
												dh = bottom - top - layer->bounds.MaxY + layer->bounds.MinY;

												MoveSizeLayer(layer, dx, dy, dw, dh);
											}
										}
										
										CheckLayers(w->WScreen, IntuitionBase);

										UnlockLayers(&w->WScreen->LayerInfo);

										if (w->WScreen->Flags & AUTOSCROLL) ScreenPosition(w->WScreen,SPOS_MAKEVISIBLE,w->LeftEdge,w->TopEdge,w->LeftEdge + w->Width - 1,w->TopEdge + w->Height - 1);
									} else {
										data->drag_refreshed = FALSE;
									}
								}
							}
						}
						else
						{
							#ifdef DELAYEDDRAG
							if ((!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE)) && (!data->drag_layerlock))
							{
								LockLayers(&w->WScreen->LayerInfo);
								data->drag_layerlock = TRUE;
							}
							#endif

								/* Rerender the window frame */
							drawwindowframe(w->WScreen
									, data->rp
									, data->curleft
									, data->curtop
									, data->curleft + w->Width  - 1
									, data->curtop  + w->Height - 1
									, IntuitionBase);
							data->isrendered = TRUE;
						}

					}

					retval = GMR_MEACTIVE;

					break;
				}

			default:
				retval = GMR_NOREUSE;
				if (!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE)) data->drag_canceled = TRUE;
				break;

			} /* switch (ie->ie_Code) */
			break;

	case IECLASS_TIMER:
		    #ifdef DELAYEDDRAG
			if ((!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE)) && (!data->drag_layerlock))
			{
				UQUAD currenttime;
				currenttime = ie->ie_TimeStamp.tv_secs;
				currenttime = currenttime * 50;
				currenttime += ie->ie_TimeStamp.tv_micro / 20000;

				if (currenttime > data->lasteventtime + 10)
				{
					LockLayers(&w->WScreen->LayerInfo);
					data->drag_layerlock = TRUE;

					drawwindowframe(w->WScreen
						, data->rp
						, data->curleft
						, data->curtop
						, data->curleft + w->Width  - 1
						, data->curtop  + w->Height - 1
						, IntuitionBase
						);
					data->lasteventtime = currenttime;
					data->isrendered = TRUE;
				}
			}
			#endif
			break;

	case IECLASS_NEWTIMER:
			if (data->movetask)
			{
				if (!data->drag_refreshed)
				{
					ih_fire_intuimessage(   w,
						  IDCMP_CHANGEWINDOW,
						  CWCODE_MOVESIZE,
						  w,
						  IntuitionBase);
					data->drag_refreshed = TRUE;
				}
				CheckLayers(w->WScreen,IntuitionBase);
			} else {
				if (MOVEHACK)
				{
					WORD newx = data->curleft - w->LeftEdge, newy = data->curtop - w->TopEdge;

					if (!data->drag_refreshed && WindowsReplied(w->WScreen,IntuitionBase))
					{
						struct Requester *req;
						ih_fire_intuimessage( w,
											  IDCMP_CHANGEWINDOW,
											  CWCODE_MOVESIZE,
											  w,
											  IntuitionBase);

						LockLayers(&w->WScreen->LayerInfo);

						w->LeftEdge    += newx;
						w->TopEdge     += newy;

						if (BLAYER(w))
						{
							MoveSizeLayer(BLAYER(w), newx, newy , 0, 0);
						}

						MoveSizeLayer(WLAYER(w), newx, newy, 0, 0);

						for (req = w->FirstRequest; req; req = req->OlderRequest)
						{
							struct Layer *layer = req->ReqLayer;

							if (layer)
							{
								int dx, dy, dw, dh;
								int left, top, right, bottom;

								left = data->curleft + req->LeftEdge;
								top = data->curtop + req->TopEdge;
								right = left + req->Width - 1;
								bottom = top + req->Height - 1;

								if (left > data->curleft + w->Width - 1)
									left = data->curleft + w->Width - 1;

								if (top > data->curtop + w->Height - 1)
									top = data->curtop + w->Height - 1;

								if (right > data->curleft + w->Width - 1)
									right = data->curleft + w->Width - 1;

								if (bottom > data->curtop + w->Height - 1)
									bottom = data->curtop + w->Height - 1;

								dx = left - layer->bounds.MinX;
								dy = top - layer->bounds.MinY;
								dw = right - left - layer->bounds.MaxX + layer->bounds.MinX;
								dh = bottom - top - layer->bounds.MaxY + layer->bounds.MinY;

								MoveSizeLayer(layer, dx, dy, dw, dh);
							}
						}

						CheckLayers(w->WScreen, IntuitionBase);

						UnlockLayers(&w->WScreen->LayerInfo);

						if (w->WScreen->Flags & AUTOSCROLL) ScreenPosition(w->WScreen,SPOS_MAKEVISIBLE,w->LeftEdge,w->TopEdge,w->LeftEdge + w->Width - 1,w->TopEdge + w->Height - 1);

						data->drag_refreshed = TRUE;
					}
				}
			}
			break;
		} /* switch (ie->ie_Class) */

	} /* if (gi) */

	return retval;
}

/***********************************************************************************/

static IPTR dragbar_goinactive(Class *cl, Object *o, struct gpGoInactive *msg)
{
	struct dragbar_data *data;
	struct Window	*w;

	data = INST_DATA(cl, o);
	w = msg->gpgi_GInfo->gi_Window;

	if (POINTERCHANGE)
	{
		IBase->GadgetPointer = 0;
		CheckMouseZone(TRUE,IntuitionBase);
		UpdateScreenPointer(w->WScreen,IntuitionBase);
	}

	/* Always clear last drawn frame */

	if (data->isrendered && data->rp && (!(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE)))
	{

		SetDrMd(data->rp, COMPLEMENT);

		/* Erase old frame */
		drawwindowframe(w->WScreen
				, data->rp
				, data->curleft
				, data->curtop
				, data->curleft + w->Width  - 1
				, data->curtop  + w->Height - 1
				, IntuitionBase
			       );

	}

	data->isrendered = FALSE;

	if (!data->drag_refreshed) CheckLayers(w->WScreen, IntuitionBase);

	Forbid();
	if (data->movetask)
	{
		Signal(data->movetask,WSIG_DIE);
		data->movetask = 0;
	}
	Permit();

	if (!data->drag_canceled)// && !(IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE))
	{
		MoveWindow(w
			   , data->curleft - w->LeftEdge	/* dx */
			   , data->curtop  - w->TopEdge	/* dy */
			  );

	}

	if (data->drag_canceled && (IBase->IControlPrefs.ic_Flags & ICF_OPAQUEMOVE))
	{
		MoveWindow(w
			   , data->startleft - w->LeftEdge /* dx */
			   , data->starttop - w->TopEdge /* dy */
			  );
	}

	if (data->drag_newtimer)
	{
		struct IOStdReq req;
		ULONG cmd = IND_SETNEWTIMER_STOP;

		memclr(&req,sizeof (struct IOStdReq));

		req.io_Device = IBase->InputIO->io_Device;
		req.io_Unit = IBase->InputIO->io_Unit;
		req.io_Command = IND_SETNEWTIMER;
		req.io_Data = &cmd;
		DoIO((struct IORequest *)&req);

		data->drag_newtimer = FALSE;
	}

	ih_fire_intuimessage(w,
						 IDCMP_CHANGEWINDOW,
						 CWCODE_MOVESIZE,
						 w,
						 IntuitionBase);

	data->drag_canceled = TRUE;

	if (data->drag_layerlock)
	{
		UnlockLayers(&w->WScreen->LayerInfo);
		data->drag_layerlock = FALSE;
	}

#ifdef USEGADGETLOCK
	if (data->drag_gadgetlock)
	{
		ReleaseSemaphore(&IBase->GadgetLock);
		data->drag_gadgetlock = FALSE;
	}
#endif

#ifdef USEWINDOWLOCK
	if (data->drag_windowlock)
	{
		UNLOCKWINDOW;
		data->drag_windowlock = FALSE;
	}
#endif

	/* User throught with drag operation. Unlock layesr and free
	rastport clone
	*/

	if (data->rp)
	{
		FreeRastPort(data->rp);
		data->rp = NULL;
	}

	return TRUE;


}

/***********************************************************************************/

/***********  Window dragbar class **********************************/

AROS_UFH3S(IPTR, dispatch_dragbarclass,
		   AROS_UFHA(Class *,  cl,  A0),
		   AROS_UFHA(Object *, o,   A2),
		   AROS_UFHA(Msg,      msg, A1)
		  )
{
	AROS_USERFUNC_INIT

	IPTR retval = 0UL;

	EnterFunc(bug("dragbar_dispatcher(mid=%d)\n", msg->MethodID));
	switch (msg->MethodID)
	{
	case GM_RENDER:
		//dragbar_render(cl, o, (struct gpRender *)msg);
		break;

	case GM_LAYOUT:
		break;

	case GM_DOMAIN:
		break;

	case GM_GOACTIVE:
		retval = dragbar_goactive(cl, o, (struct gpInput *)msg);
		break;

	case GM_GOINACTIVE:
		retval = dragbar_goinactive(cl, o, (struct gpGoInactive *)msg);
		break;

	case GM_HANDLEINPUT:
		retval = dragbar_handleinput(cl, o, (struct gpInput *)msg);
		break;

	case GM_HITTEST:
		retval = 1;
		break;

	case GM_MOVETEST:
		retval = dragbar_movetest(cl, o, (struct gpInput *)msg);
		break;

	case OM_NEW:
		retval = DoSuperMethodA(cl, o, msg);
		if (retval)
		{
			((struct Gadget *)retval)->GadgetType |= GTYP_SYSGADGET | GTYP_WDRAGGING;
		}
		break;


	default:
		retval = DoSuperMethodA(cl, o, msg);
		break;
	}

	ReturnPtr ("dragbar_dispatcher", IPTR, retval);

	AROS_USERFUNC_EXIT
}

/***********************************************************************************/

#undef IntuitionBase

/***********************************************************************************/

struct IClass *InitDragBarClass (struct IntuitionBase * IntuitionBase)
{
	struct IClass *cl = NULL;

	/* This is the code to make the dragbarclass...
	*/
	if ( (cl = MakeClass(NULL, GADGETCLASS, NULL, sizeof (struct dragbar_data), 0)) )
	{
		cl->cl_Dispatcher.h_Entry    = (APTR)AROS_ASMSYMNAME(dispatch_dragbarclass);
		cl->cl_Dispatcher.h_SubEntry = NULL;
		cl->cl_UserData		     = (IPTR)IntuitionBase;

	}

	return (cl);
}

